######################################################################
#:CQCC::IOPerlTK
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2008. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

=head1 NAME 

CQCC::IOPerlTK - PerlTk based subclass of IOBase for user interaction

=head1 CLASS DESCRIPTION

This class is an IOBase subclass which provides an set of user
interaction dialog methods (IOPrompt, IOMessage, IOListSelect, and
IOListTreeSelect) using the PerkTK widget toolkit instead of a text
menu.

It also provides access to a small set of wrapper objects based on the
IOPtkWidget class that encapsulate Perl/TK code to provide building
blocks for the CQCC Graphical User Interface. These building blocks
provide data interactors (IOPtkManip), menu support (IOPtkMenu), and
panels (IOPtkPanel) such as IOFrames or dialogs to build the GUI.

Isolating the Perl/TK code within these wrappers isolates the
Trigger class from Perl/TK dependencies, standardizes GUI components,
hides implementation details, and provides basic services such as
message catalog translations needed by the integration.

The IOPerlTK object provides methods to build and manage named
instances of the IOPtkWidget classes without needing to save the 
pointers in the Trigger methods. 

See the TriggerCQCC::GUIBuild method for an extended example of 
how this intermediate layer works. A small example shows the basic
features:

    # Get a handle to the IOPerlTK instance. 
    my $IO = &IOstd();

    # The Panel method creates a panel of a given subtype such as 
    # an IOMainWindow, IOFrame, or IODialogBox. Attributes such as 
    # TITLE are passed in to modify the instance. All IOPerlTK 
    # attributes are in CAPS without dashes; Perl/Tk attributes are 
    # lower case with dashes. 
    #
    # The TK attribute is a general pass through mechanism to pass 
    # in any attributes the internal Perl/TK widget will recognize. 
    #
    # The PACK attribute will be passed through to the Perl/Tk pack
    # operation. 

    $IO->Panel("Main", "", "IOMainWindow");
    $IO->Panel("Frame1", "Main", "IOFrame",
               TITLE=>"Frame1",
               TK=>[-borderwidth => 2, -relief => 'groove'],
               PACK=>[-side => "left", -fill => 'y']);

    # The Menu method creates a menu of command buttons and
    # sets up their callbacks. The MenuManageButtons method allows
    # access to disabling buttons as needed. 

    $IO->Menu("MyMenu", "Frame1", "IOCommandMenu", 
               TITLE=>"Menu Demo", 
               MENULIST => ["OK "    => 'print "OK\n";',
                            "Cancel" => 'print "Cancel\n";']);

    # Manipulators manage labels or buttons associated with Parm
    # objects or values. The ManipManageValues method is used to 
    # get or set ranges and values. 

    $IO->Manipulator("Label1", "Frame1", "IOLabel",
               TITLE=>"Label1", PARM=>"CQCC_CTRL_C");

    $IO->Manipulator("List1", "Frame1", "IOListbox",
               TITLE=>"List", 
               RANGE => ["a","b","c"],
               PACK=>[-side => "left", -fill => 'both']);

The IOPerlTK package was written to suit the immediate needs of the
integration GUI and is not very general. Extensions can be made as
needed by subclassing the IOPerlTK class to modify the existing
Manipulator, Menu, or Panel methods to use subclasses of the original
IOPtk* classes if needed and adding additional subtype initializer
methods.

Direct access to the underlying Perl/Tk widgets can be obtained using
the CQCC::IOPtkWidget::SubWidget method but this should be used very
carefully.

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package declarations
######################################################################

require 5.006; 
package CQCC::IOPerlTK;

use Tk 800.022;
use Tk::DialogBox;
use CQCC::IOBase;
@ISA = qw( CQCC::IOBase );
             
use strict;

use CQCC::Parm;
use CQCC::MsgCat_English;
use CQCC::IOPtkWidget;
use CQCC::IOPtkManip;
use CQCC::IOPtkMenu;
use CQCC::IOPtkPanel;

sub Version { return '2.1' };

######################################################################
#::Class Globals 
######################################################################

my $logWarning = 0;
my ($_CallbackObject,      # Object to be called on exit
    $_CallbackMethod,      # Method to be called on exit
    $_IdleCallbackHandler, # Handler returned by PerlTK
    $_MainWindow,          # TODO: Cleanup how this is remembered between
                           # setting and handling callback
   );

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(@args)>

This class function sets IOBaseInterface to make an instance of 
this subclass then runs the class test for IOBase.

IOPtkPanel::ClassTest is a local test of the IOPtkWidget based
widgets. 

=back

=cut

######################################################################

sub ClassTest { 
    my @args = @_;
    &ForceParm("IOBaseInterface", "CQCC::IOPerlTK");
    return &CQCC::IOBase::ClassTest(@args);
}

######################################################################
#::method Shutdown

=over 6

=item C<Shutdown(status)>

This method will hide and delete the main windows before calling the main
exit handler. 

=back

=cut 

######################################################################

sub Shutdown {
    my ($self, $status) = @_;
    $self->ShowMainWindow("HIDE");
    if (defined $self->{MAIN}) {
        $self->{MAIN}->destroy();
    }
    CQCC::IOBase::Shutdown($self,$status);
    return;
}

######################################################################
#::method GetMainWindow

=over 6

=item C<GetMainWindow()>

This method returns a pointer to the Perk/Tk MainWindow widget;
if it has not been created yet, it is created with default 
properties and returned. 

=back

=cut 

######################################################################

sub GetMainWindow {
    my ($self) = @_;
    if (! defined $self->{MAIN}) {
        # Command to create MainWindow panel will call SetMainWindow
        $self->Panel("TempMain", "", "IOMainWindow");
    }
    return &IOstd()->{MAIN};
}

######################################################################
#::method ListSelect

=over 6

=item C<ListSelect(multiChoice, titleMsgTag, choiceMsgTagList...)>

This method presents a menu with a title described by "titleMsgTag"
and one or more options to select from described by the remaining
arguments. If multiChoice is 1 then multiple choices are allowed, 
otherwise only one choice is provided.

In the PerlTK implementation, the choices are displayed in a scrolling
"list" type menu and the user is expected to pick one or more lines in
the list.

The selected items will be returned as a list to the caller, 
using the original choiceMsgTag items in their entirety for the caller
to match to expectations. 

The return value consists of a list containing a flag to abort
(1 for abort, 0 for continue) and the answer(s).

This method should be called through the IOBase::IOListSelect 
function. 

=back 

=cut

######################################################################

sub ListSelect {
    my ($self, $multiChoice, $titleMsgTag, @choices) = @_;

    my (@answers,$i,$done,$choice, $msg, $menuLine,$maxItems);
    my (@displayChoices, $indexedAnswer, $abort);
    
    $self->_LogEntry("ListSelect", $titleMsgTag);

    $indexedAnswer = $self->_ReadSTDIN( "BEFORE",
           ["IO_P_LISTSELECT", ($multiChoice ? "(s)" : "")]);

    ########################################### IOPerlTk code
    # Create a top level Dialog Box with a text label
    # and a scrollable list area. 

    my $geometry = $self->{LIST_STYLE} eq "IOListboxTree" ? 
                    $self->{LISTTREE_GEOM} : $self->{LIST_GEOM};
    my @dialogButtons = ( ["GUI_Commit"], ["GUI_Cancel"]);
    my $dialogBox = $self->Panel("ListDialogBox", "", "IODialogBox",
                                 BUTTONS => \@dialogButtons,
                                 GEOMETRY => $geometry,
                                 TITLE => [ "GUI_ListSelect" ]);
    if ($titleMsgTag) { 
        $self->Manipulator("ListDialogLabel", "ListDialogBox",
                           "IOLabel",
                           TK => [-justify => 'left'],
                           PACK => [-side => "top", -anchor => "w"],
                           VALUE => $titleMsgTag);
    }

    # Creates an IOListbox or IOListboxTree as needed. 

    $self->Manipulator("ListDialogList", "ListDialogBox",
                       $self->{LIST_STYLE}, 
                       MODE => ($multiChoice ? "browse" : "single"),
                       # Separator ignored for regular listbox
                       SEPARATOR => $self->{LIST_SEP},
                       # DIALOG values only used by ListboxTree
                       DIALOG_BOX => "ListDialogBox",
                       DIALOG_BUTTON => [ "GUI_Commit" ],
                       PACK => [-side => "top", -anchor => "w", 
                                -expand => 1, -fill => "both"]);
                       
    ##################################################
    # Prep choices, handle user interaction, 
    # then translate answers into numbers for recording to journal. 

    if (! defined $indexedAnswer) { 
        # Translate messages
        for (@choices) {
            push @displayChoices, $self->{MSGCAT}->Message($_);
        }

        ######################################### IOPerlTk code
        $self->ManipManageValues("ListDialogList", "SETRANGE",
                                  \@displayChoices);
        $abort = $dialogBox->IODialogBoxShow();
        @answers = $self->ManipManageValues("ListDialogList",
                                            "GETVALUE");
        ########################################
         
        # Need to translate answers into numbers on a single string
        # in order to store the answer and need to translate the 
        # modified choice strings (stripped commas, etc) to match 
        # original choices.
        # Return the original choices list not the modified version.
        # Strip trailing spaces off both the answer and the test
        #   strings because ClearPrompt may strip off spaces

        my $numberString = "";
        my (@realAnswers, $displayAnswer, $status);

        for $displayAnswer (@answers) { 

            # to compare with $choice in the format
            # Don't expect conversion error since we just converted
            # these the other way to set up the list. 
            ($status, $displayAnswer) = 
                &IOConvertString($displayAnswer, "CORE_CODEPAGE", 
                                 "IO_CODEPAGE");
            my $i = 1;
            $displayAnswer =~ s/\s+$//;
            for (@displayChoices) {
                my $choice = $_;
                $choice =~ s/\s+$//;
                if ($displayAnswer eq $choice) { 
                   # Note: indices are 1-based
                   push @realAnswers, $choices[$i-1];
                   $numberString .= "$i ";
                   last;
                }
                $i++;
            } 
        }
        @answers = @realAnswers;
        $indexedAnswer = $numberString;

    ##################################################
    # If log file provided answer string then translate numbers 
    # back into choices (without comma stripping, etc)

    } else { 
        my @numberedChoices = &IOParseListString($indexedAnswer);
        my @realAnswers;
        
        for (@numberedChoices) {
            my $index = $_;   # Note: indices are 1-based
            if (($index > 0) && ($index <= $#choices)) {
                push @realAnswers, $choices[$index-1];
            }
            else {
                &IOPrint("Error: IOPerlTK::ListSelect journal " .
                    "answer list index ($index) out of range\n");
            }
        }
    }

    ########################################## IOPerlTK code: 
    if ($dialogBox) { 
        $dialogBox->DESTROY();
    } 
    ########################################

    # Save the numbered string so that the text driven playback 
    # can work with the IOBase version of ListSelect().

    $self->_ReadSTDIN( "AFTER",
        ["IO_P_LISTSELECT", ($multiChoice ? "(s)" : "")],
        $indexedAnswer);

    return ($abort, @answers);
}

######################################################################
#::method ListTreeSelect

=over 6

=item C<ListTreeSelect(multiChoice,separator,titleMsgTag,
choiceMsgTagList...)>

Variant on ListSelect which presents the list as a folding tree
structure based on the choice list, assuming that each choice is a
pathname separated by the separator character.

If the IOBase class does not provide support for the IOListTreeSelect
method, it will default to using the standard IOListSelect method
instead and present the information as a normal list.

For more details see the IOListSelect function documentation.

This method should be called through the IOBase::IOListTreeSelect 
function.

=back 

=cut


######################################################################

sub ListTreeSelect {
    my ($self, $multiChoice, $separator, $titleMsgTag, 
        @choiceMsgTagList)=@_;
    my @saveState = ($self->{LIST_STYLE}, $self->{LIST_SEP});
    $self->{LIST_STYLE} = "IOListboxTree";
    $self->{LIST_SEP} = $separator;
    my ($abort, @answers) =
        $self->ListSelect($multiChoice, $titleMsgTag, 
                          @choiceMsgTagList);
    ($self->{LIST_STYLE}, $self->{LIST_SEP}) = @saveState;
    return ($abort, @answers);
}

######################################################################
#::method Manipulator

=over 6

=item C<Manipulator(name, parent, subtype, key=>value,...)>

This method creates a new IOPtkManip object to provide a GUI
component. See the documentation for CQCC::IOPtkManip for more
information on the available subtypes and arguments. 

Use the CQCC::IOPerlTK::ManipManageValues method to modify properties
of the manipulator or to retrieve value or range information. 

=back

=cut 

######################################################################

sub Manipulator{
    my ($self,@args) = @_;
    return new CQCC::IOPtkManip(@args);
}

######################################################################
#::method ManipManageValues

=over 6

=item C<ManipManageValues(name, operation, @arglist)>

This method is called to get or set the value or range of a named
manipulator widget. The currently supported operations include the
following:

    GETRANGE - retrieves widget's current range as a list reference
    SETRANGE - sets widget range to a list reference value
    GETVALUE - retrieves widget's current range as a list
    SETVALUE - sets widget value to list reference value

The value argument is only expected by the SET* operations. 

See the documentation for CQCC::IOPtkManip::ManageValues for the most
recent information on the available subtypes and arguments.

=back

=cut 

######################################################################

sub ManipManageValues{
    my ($self, $name, $op, @args) = @_;
    my $widget= &WidgetLookup($name);
    return $widget->ManageValues($op, @args);
}

######################################################################
#::method Menu

=over 6

=item C<Menu(name, parent, subtype, key=>value,...)>

This method creates a new IOPtkMenu object to provide a GUI menu
component. See the documentation for CQCC::IOPtkMenu for more
information on the available subtypes and arguments.

Use the CQCC::IOPerlTK::MenuManageButtons method to modify the
state of the menu buttons. 

=back

=cut 

######################################################################

sub Menu{
    my ($self,@args) = @_;
    return new CQCC::IOPtkMenu(@args);
}

######################################################################
#::method MenuManageButtons

=over 6

=item C<MenuManageButtons(name, buttonName, value)>

This method can be called for any IOPtkMenu widget to modify the state
of its buttons. Operations currently supported include the following:
STATE which changes the buttonName's state to value (normal, disabled,
or active).

See the documentation for CQCC::IOPtkMenu::ManageButtons for the most
recent information on the available subtypes and arguments.

=back

=cut 

######################################################################

sub MenuManageButtons{
    my ($self,$menuName,@args) = @_;
    my $widget = &WidgetLookup($menuName);
    return $widget->ManageButtons(@args);
}

######################################################################
#::method Message

=over 6

=item C<Message(type, abort, [msgtag, args])>

This method (called through IOMessage()) will display its message
to the user, after translating the msgtag and its arguments into the
corresponding message catalog entry. 

The type value may be "INFO", "WARN", "ERROR", or "FATAL" and is
intended to qualify how the message is to be shown and whether or
not the message affects the overall trigger status. If "FATAL" the
status will be set to failed and the trigger will try to cleanly
unwind and report failure back to the original caller, usually
ClearCase.

The IOPerlTK implementation provides a resizeable scrolling message
box for large messages over SCROLL_THRESHOLD lines.

The abort flag is also used to signal general failure for terminal
conditions and will be combined with the type information shortly. 

=back 

=cut

######################################################################

sub Message {
    my ($self, $type, $msgTag) = @_;
    my ($package, $filename, $line) = caller 1;
    my $proceed = 1;
    my $answer;

    if (! defined $msgTag) {
        &IOMessage("ERROR", ["IO_E_MSGTAG_UNDEF"]);
        IOTraceBack("Message", 10);
        return;
    }

    my $msgType = $self->{MSGCAT}->Message([$type]);
    my $msg = $self->{MSGCAT}->Message($msgTag) . "\n";
    $self->_LogEntry("Message", $msgTag);
    
    my $displayMsg = $msgType . ": " . $msg;

    if ($type eq "ERROR" || $type eq "FATAL") {
        $displayMsg .= sprintf "Caller: $filename Line: $line\n";
    }

    $answer = $self->_ReadSTDIN("BEFORE", ["IO_P_MESSAGE" ]);
    if (! defined $answer) { 
         my @lines = split("\n",$displayMsg); # Count lines
         my $buttonOk = $self->{MSGCAT}->Message( ["GUI_Commit"] );
         
         ########################################
         # Perl/TK code
         if ($#lines <= $self->{SCROLL_THRESHOLD}) {
             my %iconMap = ( "INFO" => "info", 
                             "WARN" => "question",
                             "ERROR" => "warning", 
                             "FATAL" => "error");
             my $title = $self->{MSGCAT}->Message( ["GUI_Message"] );
             # messageBox doesn't take geometry 
             my $result = $self->GetMainWindow()->
                 messageBox(-icon    => $iconMap{$type},
                            -title   => $title,
                            -message => $displayMsg,
                            -type    => $buttonOk);
         } else {
             my $dialogBox = 
                 $self->Panel("MsgDialogBox", "", "IODialogBox",
                              GEOMETRY => $self->{MESSAGE2_GEOM},
                              BUTTONS => [ $buttonOk ],
                              TITLE => [ "GUI_Message" ]);
             my $sLabel = 
                 $self->Manipulator("MsgDialogLabel", "MsgDialogBox",
                                    "IOScrolledLabel",
                                    TK => [-justify => 'left'],
                                    PACK => [-side => "top",
                                             -anchor => "w",
                                             -expand => 1, 
                                             -fill => 'both'],
                                    VALUE => $displayMsg);
             my $abort = $dialogBox->IODialogBoxShow();
             $dialogBox->DESTROY();
         }
         ########################################
    }    
    $self->_ReadSTDIN("AFTER", ["IO_P_MESSAGE" ], $answer);
    
    # Record message to error log if enabled
    $self->MessageLog($type,$msgTag,$filename,$line);

    if ($type eq "FATAL") {
        &IOExitHandler(2);
    } elsif ($proceed == 0) {
        &IOExitHandler(1);
    }

    return;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::IOPerlTK()>

This method initializes the object's attributes.

=back

=cut

######################################################################

sub new { 
   my $proto = shift;
   my $class = ref($proto) || $proto;  # Object or class name

   # Attributes
   my $self;

   # CCperl doesn't support SUPER so use direct reference to parent
   # IOBase is expected to complete initialization before returning
   # which should also preset file pathname suffixes, etc. 
   # Still can't call IO* functions until _IOStd is created and
   # we are still in the middle of that when this method is called. 

   &CQCC::MsgCat::MsgCat_Prompts("GUI", "Yes", "Abort");

   $self = new CQCC::IOBase();

   # Won't be communicated to MsgCat until Initialize routine
   $self->{IO_CODEPAGE} = "UTF8_CODEPAGE";

   $self->{CLASSNAME} = "CQCC::IOPerlTK";
   $self->{LIST_STYLE} = "IOListbox";
   $self->{LIST_SEP} = "/";
   $self->{SCROLL_THRESHOLD} = 15;
   $self->{LIST_GEOM} = "350x300";
   $self->{LISTTREE_GEOM} = "350x300";
   #Not used $self->{MESSAGE1_GEOM};
   $self->{MESSAGE2_GEOM} = "450x400";
   #Not used $self->{PROMPT_GEOM};

   bless($self, $class);
   return $self;
}

######################################################################
#::method Panel

=over 6

=item C<Panel(name, parent, subtype, key=>value,...)>

This method creates a new IOPtkPanel object to provide a GUI geometry
component. See the documentation for CQCC::IOPtkPanel for more
information on the available subtypes and arguments. 

=back

=cut 

######################################################################

sub Panel{
    my ($self,@args) = @_;
    my $widget = new CQCC::IOPtkPanel(@args);
    return $widget;
}

######################################################################
#::method Prompt

=over 6

=item C<Prompt(promptMsg, optionsList)>

This internal function displays its message to the user, after
translating the msgtag and its arguments into the corresponding
message catalog entry. In addition to the required message, it 
accepts optional qualifiers in the form of "option=>value" pairs to
support multiple public API calls like IOPromptLogin.

Available options include the following: 

   "IOPROMPT_MODE" =>     # Prompt display mode, one of the following

       "IOPROMPT_TEXT"   # Default, display message, expect user input 

  or   "IOPROMPT_LOGIN"  # Login prompt, expect user to enter username and (masked) password

  or   "IOPROMPT_YES_NO" # Present user choice of yes or no input

   "IOPROMPT_LOGIN_MSG"=>loginPromptMessage  # Alternate login prompt

The return value consists of a list containing the abort status
(1 for abort, 0 for continue) and the answer(s). 

"IOPROMPT_LOGIN" returns two strings, username and password.
"IOPROMPT_YES_NO" returns a single integer, 1 for yes, 0 for no.

($abort, $answer(s)) = $obj->Prompt(msg, option1=>value1, ... );

=back

=cut

######################################################################

sub Prompt {
    my ($self, $msgTag, %options) = @_;
    my $mode = (defined $options{"IOPROMPT_MODE"}) ? 
               $options{"IOPROMPT_MODE"} : "IOPROMPT_TEXT";

    my ($answer, @answerList, $abort);
    my ($nameStr, $passStr, $name, $password);
    if ($mode eq "IOPROMPT_LOGIN") { 
        if (defined $options{"IOPROMPT_LOGIN_MSG"}) {
            $msgTag = $options{"IOPROMPT_LOGIN_MSG"};
        }
        $nameStr = $self->{MSGCAT}->Message( ["IO_P_USERNAME"]);
        $passStr = $self->{MSGCAT}->Message( ["IO_P_PASSWORD"]);
    }
    my $msg = $self->{MSGCAT}->Message($msgTag);

    $self->_LogEntry("Prompt", $msgTag);

    $answer = $self->_ReadSTDIN("BEFORE", ["IO_P_MESSAGE" ]);

    ########################################
    # Perl/TK code for DialogBox

    my $titleStr = $self->{MSGCAT}->Message( ["GUI_Prompt"]);
    my ($buttonOk, $buttonYes, $buttonNo, $buttonAbort);
    my @buttonList;

    if ($mode eq "IOPROMPT_YES_NO") {
        $buttonYes = $self->{MSGCAT}->Message( ["IO_P_YES"]);
        $buttonNo = $self->{MSGCAT}->Message( ["IO_P_NO"]);
        push(@buttonList, $buttonYes, $buttonNo);
    } else {
        $buttonOk = $self->{MSGCAT}->Message( ["GUI_Commit"]);
        push(@buttonList, $buttonOk);
    }
    $buttonAbort = $self->{MSGCAT}->Message( ["GUI_Cancel"]);
    push(@buttonList, $buttonAbort);
    my $d = $self->GetMainWindow()->
         DialogBox(-title => $titleStr,
                   -buttons => [ @buttonList ]);
    my @lines = split("\n", $msg);
    my $minLen = 20;
    my $dialogBoxInput;

    for (@lines) { 
        if (length($_) > $minLen) { 
            $minLen = length($_);
        } 
    } 
    my $w = $d->add("Label",
                    -justify => 'left',
                    -text=> $msg)->pack(-padx=>10, -pady=>10);
    if ($mode eq "IOPROMPT_LOGIN") {
        my $nameFrame = $d->Frame(
                        )->pack(-fill => "both", -expand=>1,
                                -padx=>10, -pady=>10);
        my $name_w = $nameFrame->Label(-justify => 'left',
                                       -text=> $nameStr
                                      )->pack(-side => 'left');
        $name_w = $nameFrame->Entry(-textvariable => \$name,
                                   )->pack(-side => "left", -expand=>1);

        my $passFrame = $d->Frame()->pack(-fill => "both", -expand=>1,
                                          -padx=>10, -pady=>10);
        my $pass_w = $passFrame->Label(-justify => 'left',
                                       -text=> $passStr
                                       )->pack(-side => 'left');
        $pass_w = $passFrame->Entry(-show => '*',
                                    -textvariable => \$password
                                    )->pack(-side => "left", -expand=>1);
    } elsif ($mode eq "IOPROMPT_TEXT") {
    $w = $d->add("Entry",
                    -width => $minLen,
                    -textvariable => \$dialogBoxInput)->pack;
    }
    my $button = $d->Show;
    if ($mode eq "IOPROMPT_TEXT") {
        push @answerList, $dialogBoxInput;
    } elsif ($mode eq "IOPROMPT_YES_NO") {
        push @answerList, ($button eq $buttonYes) ? 1 : 0;
    } elsif ($mode eq "IOPROMPT_LOGIN") {
        push @answerList, $name, ($password ? $password : "");
    }
    $abort = ($button eq $buttonAbort) ? 1 : 0;
    if (Exists $d) { 
        $d->destroy();
    } 
    ########################################

    $self->_ReadSTDIN("AFTER", ["IO_P_MESSAGE" ], $answer);

    my $status;
    ($status,@answerList) = &IOConvertString(\@answerList, "IO_CODEPAGE", "CORE_CODEPAGE");
    # If conversion error on user input, report it and then prompt again to let them change
    # input or abort.
    if ($status) {
        IOMessage("ERROR", 
            ["IOPTK_E_CONVERSION_ERROR", "IOPerlTK::Prompt", $status]);
        return $self->Prompt(@_);
    }

    return ($abort, @answerList);
}

######################################################################
#::method RunMainLoop

=over 6

=item C<RunMainLoop()>

This method starts the Perl/Tk MainLoop function to begin interaction
with the user after the GUI has been defined. 

=back

=cut 

######################################################################

sub RunMainLoop {
    my ($self) = @_;
    MainLoop();
    return;
}

######################################################################
#::method SetIdleCallback

=over 6

=item C<SetIdleCallback(callback)>

This method sets a callback to be invoked once the main loop is activated. 

=back

=cut 

######################################################################

sub SetIdleCallback {
    my ($self, $object, $method) = @_;

    $_CallbackMethod = $method;
    $_CallbackObject = $object;
    $_MainWindow = $self->GetMainWindow();   
    $_IdleCallbackHandler = 
        $_MainWindow->afterIdle(\&CQCC::IOPerlTK::IdleCallbackHandler);
}

sub IdleCallbackHandler {
    my ($self) = @_;
    my $exitObject = $_CallbackObject;
    my $exitMethod = $_CallbackMethod; 
    
    if ($_IdleCallbackHandler) {
        $_MainWindow->afterCancel($_IdleCallbackHandler);
        $_IdleCallbackHandler = undef;
    }

    # Make sure the main window is displayed
    $_MainWindow->raise(); 

    if ($_CallbackObject) {
       my $cmd;
       $cmd = "\$exitObject->$exitMethod()";
       eval $cmd;
    }
    return;
}

######################################################################
#::method SetMainWindow

=over 6

=item C<SetMainWindow(window)>

This method sets the Perk/Tk MainWindow widget and returns the widget
that is current MainWindow.

=back

=cut 

######################################################################

sub SetMainWindow {
    my ($self, $window) = @_;
    $self->{MAIN} = $window;
    return $self->{MAIN};
}

######################################################################
#::method ShowMainWindow

=over 6

=item C<ShowMainWindow(status)>

This method will show the main window if the status is "SHOW"
and hide it if the status is "HIDE". By default the window is 
shown. 

=back

=cut 

######################################################################

sub ShowMainWindow {
    my ($self, $status) = @_;
    if (defined $self->{MAIN}) {
        if ($status eq "SHOW") {
            $self->{MAIN}->deiconify();
            $self->{MAIN}->raise();
        }
        elsif ($status eq "HIDE") {
            $self->{MAIN}->withdraw();
        }
    }
    return;
}

######################################################################
#::Tail
######################################################################

1;
